@@ -3,17 +3,16 @@ package Convert::Pheno;
 use strict;
 use warnings;
 use autodie;
-use feature               qw(say);
+use feature qw(say);
 use File::Spec::Functions qw(catdir catfile);
 use Data::Dumper;
 use Path::Tiny;
 use File::Basename;
 use File::ShareDir::ProjectDistDir;
 use List::Util qw(any uniq);
-use Carp       qw(confess);
 use XML::Fast;
 use Moo;
-use Types::Standard                qw(Str Int Num Enum ArrayRef Undef);
+use Types::Standard qw(Str Int Num Enum ArrayRef Undef);
 use File::ShareDir::ProjectDistDir qw(dist_dir);
 
 #use Devel::Size     qw(size total_size);
@@ -21,7 +20,8 @@ use Convert::Pheno::IO::CSVHandler;
 use Convert::Pheno::IO::FileIO;
 use Convert::Pheno::SQLite;
 use Convert::Pheno::Mapping;
-use Convert::Pheno::CSV qw(do_bff2csv do_pxf2csv);
+use Convert::Pheno::CSV;
+use Convert::Pheno::RDF qw(do_bff2jsonld do_pxf2jsonld);
 use Convert::Pheno::OMOP;
 use Convert::Pheno::PXF;
 use Convert::Pheno::BFF;
@@ -41,9 +41,13 @@ $SIG{__WARN__} = sub { warn "Warn: ", @_ };
 $SIG{__DIE__}  = sub { die "Error: ", @_ };
 
 # Global variables:
-our $VERSION   = '0.17_1';
+our $VERSION   = '0.24_1';
 our $share_dir = dist_dir('Convert-Pheno');
 
+# SQLite database
+my @all_sqlites       = qw(ncit icd10 ohdsi cdisc omim hpo);
+my @non_ohdsi_sqlites = qw(ncit icd10 cdisc omim hpo);
+
 ############################################
 # Start declaring attributes for the class #
 ############################################
@@ -156,6 +160,15 @@ has [qw /data method/] => ( is => 'rw' );
 # End declaring attributes for the class #
 ##########################################
 
+sub BUILD {
+
+    # BUILD: is an instance method that is called after the object has been constructed but before it is returned to the caller.
+    # BUILDARGS is a class method that is responsible for processing the arguments passed to the constructor (new) and returning a hash reference of attributes that will be used to initialize the object.
+    my $self = shift;
+    $self->{databases} =
+      $self->{ohdsi_db} ? \@all_sqlites : \@non_ohdsi_sqlites;
+}
+
 # NB: In general, we'll only display terms that exist and have content
 
 #############
@@ -166,8 +179,9 @@ has [qw /data method/] => ( is => 'rw' );
 
 sub bff2pxf {
 
+    my $self = shift;
     # <array_dispatcher> will deal with JSON arrays
-    return array_dispatcher(shift);
+    return $self->array_dispatcher;
 }
 
 #############
@@ -178,8 +192,9 @@ sub bff2pxf {
 
 sub bff2csv {
 
+    my $self = shift;
     # <array_dispatcher> will deal with JSON arrays
-    return array_dispatcher(shift);
+    return $self->array_dispatcher;
 }
 
 #############
@@ -190,8 +205,21 @@ sub bff2csv {
 
 sub bff2jsonf {
 
+    my $self = shift;
     # <array_dispatcher> will deal with JSON arrays
-    return array_dispatcher(shift);
+    return $self->array_dispatcher;
+}
+##############
+##############
+# BFF2JSONLD #
+##############
+##############
+
+sub bff2jsonld {
+
+    my $self = shift;
+    # <array_dispatcher> will deal with JSON arrays
+    return $self->array_dispatcher;
 }
 
 ################
@@ -205,26 +233,29 @@ sub redcap2bff {
     my $self = shift;
 
     # Read and load data from REDCap export
-    my $data = read_csv( { in => $self->{in_file}, sep => undef } );
-    my ( $data_redcap_dict, $data_mapping_file ) =
-      read_redcap_dict_and_mapping_file(
+    my $data = read_csv( { in => $self->{in_file}, sep => $self->{sep} } );
+    my $data_redcap_dict = read_redcap_dict_file(
+        {
+            redcap_dictionary => $self->{redcap_dictionary},
+        }
+    );
+    my $data_mapping_file = read_mapping_file(
         {
-            redcap_dictionary    => $self->{redcap_dictionary},
             mapping_file         => $self->{mapping_file},
             self_validate_schema => $self->{self_validate_schema},
             schema_file          => $self->{schema_file}
         }
-      );
+    );
 
     # Load data in $self
-    $self->{data}              = $data;                 # Dynamically adding attributes (setter)
-    $self->{data_redcap_dict}  = $data_redcap_dict;     # Dynamically adding attributes (setter)
-    $self->{data_mapping_file} = $data_mapping_file;    # Dynamically adding attributes (setter)
-    $self->{metaData}          = get_metaData($self);
-    $self->{convertPheno}      = get_info($self);
+    $self->{data}              = $data;                  # Dynamically adding attributes (setter)
+    $self->{data_redcap_dict}  = $data_redcap_dict;      # Dynamically adding attributes (setter)
+    $self->{data_mapping_file} = $data_mapping_file;     # Dynamically adding attributes (setter)
+    $self->{metaData}          = get_metaData($self);    # Dynamically adding attributes (setter)
+    $self->{convertPheno}      = get_info($self);        # Dynamically adding attributes (setter)
 
     # array_dispatcher will deal with JSON arrays
-    return array_dispatcher($self);
+    return $self->array_dispatcher;
 }
 
 ################
@@ -247,7 +278,7 @@ sub redcap2pxf {
     $self->{in_textfile} = 0;            # setter
 
     # Run second iteration
-    return array_dispatcher($self);
+    return $self->array_dispatcher;
 }
 
 ##############
@@ -288,6 +319,7 @@ sub omop2bff {
     #   --stream
     #   Note: BFF / PXF JSON files serve as intermediate stages. They group data by individual for easier inspection but are ultimately stored in Mongo DB.
     #   Similar to the genomicVariations issue in B2RI, multiple JSON objects (like MEASUREMENTS, DRUGS) can correspond to a single individual.
+    #   The link is the term "id"
     #
     #   Potential Issues and Solutions:
     #     1. Mandatory <CONCEPT> Table:
@@ -337,15 +369,20 @@ sub omop2bff {
 
         for my $file ( @{ $self->{in_files} } ) {
             my ( $table_name, undef, $ext ) = fileparse( $file, @exts );
+
+            #####################
+            # PostgreSQL export #
+            #####################
+
             if ( $ext =~ m/\.sql/i ) {
 
-                #######################
-                # Loading OMOP tables #
-                #######################
+                print "> Param: --max-lines-sql = $self->{max_lines_sql}\n" if $self->{verbose};
 
                 # --no-stream
                 if ( !$self->{stream} ) {
 
+                    print "> Mode : --no-stream\n\n" if $self->{verbose};
+
                     # We read all tables in memory
                     $data = read_sqldump( { in => $file, self => $self } );
 
@@ -356,18 +393,28 @@ sub omop2bff {
                 # --stream
                 else {
 
+                    print "> Mode : --stream\n\n" if $self->{verbose};
+
                     # We'll ONLY load @stream_ram_memory_tables
                     # in RAM and the other tables as $fh
                     $self->{omop_tables} = [@stream_ram_memory_tables];    # setter
                     $data = read_sqldump( { in => $file, self => $self } );
                 }
 
+                # Misc print
+                print "> Parameter --max-lines-sql set to: $self->{max_lines_sql}\n\n" if $self->{verbose};
+
                 # We keep the filepath for later
                 $filepath = $file;
 
                 # Exit loop
                 last;
             }
+
+            #############
+            # CSV files #
+            #############
+
             else {
 
                 # We'll load all OMOP tables that the user is providing as -iomop
@@ -378,21 +425,24 @@ sub omop2bff {
                   #unless (any { $_ eq $table_name } @{ $omop_main_table->{$omop_version} };
                   unless any { $_ eq $table_name } @omop_essential_tables;    # global
 
+                my $msg =
+                  "Reading <$table_name> and storing it in RAM memory...";
+
                 # --no-stream
                 if ( !$self->{stream} ) {
 
                     # We read all tables in memory
+                    say $msg if ( $self->{verbose} || $self->{debug} );
                     $data->{$table_name} =
-                      read_csv( { in => $file, sep => $self->{sep} } );
+                      read_csv( { in => $file, sep => $self->{sep}, self => $self } );
                 }
 
                 # --stream
                 else {
-                    # We'll ONLY load @stream_ram_memory_tables
-                    # in RAM and the other tables as $fh
                     if ( any { $_ eq $table_name } @stream_ram_memory_tables ) {
+                        say $msg if ( $self->{verbose} || $self->{debug} );
                         $data->{$table_name} =
-                          read_csv( { in => $file, sep => $self->{sep} } );
+                          read_csv( { in => $file, sep => $self->{sep}, self => $self } );
                     }
                     else {
                         push @filepaths, $file;
@@ -405,28 +455,40 @@ sub omop2bff {
     #print Dumper_concise($data) and die;
     #print Dumper_concise($self) and die;
 
-    # Primarily with CSVs, it can happen that user does not provide <CONCEPT.csv>
-    confess 'We could not find table <CONCEPT> from your input files'
+    # *** IMPORTANT ***
+    # ABOUT TABLE <CONCEPT> BEING MANDATORY
+    # Options:
+    # a) MANDATORY <== IMPLEMENTED
+    #    Drawback: High RAM usage with <PERSON> for large tables
+    # b) OPTIONAL (with --ohdsi-db)
+    #    Using external SQLite database is possible, but risks missing ad hoc concept_ids
+    # Note: CSV users might not provide <CONCEPT.csv> with CSVs
+    die "The table <CONCEPT> is missing from the input files\n"
       unless exists $data->{CONCEPT};
 
     # We create a dictionary for $data->{CONCEPT}
-    $self->{data_ohdsi_dic} = transpose_ohdsi_dictionary( $data->{CONCEPT} );  # Dynamically adding attributes (setter)
+    $self->{data_ohdsi_dict} = convert_table_aoh_to_hoh( $data, 'CONCEPT', $self );   # Dynamically adding attributes (setter)
+
+    # Transform Array of Hashes (AoH) to Hash of Hashes (HoH) for faster computation
+    if ( $self->{stream} ) {
+        $self->{person} = convert_table_aoh_to_hoh( $data, 'PERSON', $self );         # Dynamically adding attributes (setter)
+    }
+
+    # We convert $self->{data}{VISIT_OCCURRENCE} if present
+    if ( exists $data->{VISIT_OCCURRENCE} ) {
+        $self->{visit_occurrence} =
+          convert_table_aoh_to_hoh( $data, 'VISIT_OCCURRENCE', $self );               # Dynamically adding attributes (setter)
+        delete $data->{VISIT_OCCURRENCE};                                      # Anyway, $data->{VISIT_OCCURRENCE} = [] from convert_table_aoh_to_hoh
+    }
 
     # We load the allowed concept_id for exposures as hashref (for --no--stream and --stream)
     $self->{exposures} = load_exposures( $self->{exposures_file} );            # Dynamically adding attributes (setter)
 
-    # We transpose $self->{data}{VISIT_OCCURRENCE} if present
-    if ( exists $data->{VISIT_OCCURRENCE} ) {
-        $self->{visit_occurrence} =
-          transpose_visit_occurrence( $data->{VISIT_OCCURRENCE} );             # Dynamically adding attributes (setter)
-        delete $data->{VISIT_OCCURRENCE};
-    }
-
-    # Now we need to perform a tranformation of the data where 'person_id' is one row of data
+    # Now we need to perform a transformation of the data where 'person_id' is one row of data
     # NB: Transformation is due ONLY IN $omop_main_table FIELDS, the rest of the tables are not used
     # The transformation is performed in --no-stream mode
     $self->{data} =
-      $self->{stream} ? $data : transpose_omop_data_structure($data);    # Dynamically adding attributes (setter)
+      $self->{stream} ? $data : transpose_omop_data_structure($self, $data);    # Dynamically adding attributes (setter)
 
     # Giving some memory back to the system
     $data = undef;
@@ -445,7 +507,7 @@ sub omop2bff {
     # --no-stream
     else {
         # array_dispatcher will deal with JSON arrays
-        return array_dispatcher($self);
+        return $self->array_dispatcher;
     }
 }
 
@@ -479,7 +541,7 @@ sub omop2pxf {
         $self->{in_textfile} = 0;            # setter
 
         # Run second iteration
-        return array_dispatcher($self);
+        return $self->array_dispatcher;
 
         # CLI
     }
@@ -508,15 +570,18 @@ sub cdisc2bff {
     my $hash = xml2hash $str, attr => '-', text => '~';
     my $data = cdisc2redcap($hash);
 
-    my ( $data_redcap_dict, $data_mapping_file ) =
-      read_redcap_dict_and_mapping_file(
+    my $data_redcap_dict = read_redcap_dict_file(
+        {
+            redcap_dictionary => $self->{redcap_dictionary},
+        }
+    );
+    my $data_mapping_file = read_mapping_file(
         {
-            redcap_dictionary    => $self->{redcap_dictionary},
             mapping_file         => $self->{mapping_file},
             self_validate_schema => $self->{self_validate_schema},
             schema_file          => $self->{schema_file}
         }
-      );
+    );
 
     # Load data in $self
     $self->{data}              = $data;                 # Dynamically adding attributes (setter)
@@ -524,7 +589,7 @@ sub cdisc2bff {
     $self->{data_mapping_file} = $data_mapping_file;    # Dynamically adding attributes (setter)
 
     # array_dispatcher will deal with JSON arrays
-    return array_dispatcher($self);
+    return $self->array_dispatcher;
 }
 
 ###############
@@ -547,7 +612,7 @@ sub cdisc2pxf {
     $self->{in_textfile} = 0;            # setter
 
     # Run second iteration
-    return array_dispatcher($self);
+    return $self->array_dispatcher;
 }
 
 #############
@@ -558,8 +623,65 @@ sub cdisc2pxf {
 
 sub pxf2bff {
 
+    my $self = shift;
     # <array_dispatcher> will deal with JSON arrays
-    return array_dispatcher(shift);
+    return $self->array_dispatcher;
+}
+
+#############
+#############
+#  CSV2BFF  #
+#############
+#############
+
+sub csv2bff {
+
+    my $self = shift;
+
+    # Read and load data from CSV
+    my $data = read_csv( { in => $self->{in_file}, sep => $self->{sep} } );
+
+    # Read and load mapping file
+    my $data_mapping_file = read_mapping_file(
+        {
+            mapping_file         => $self->{mapping_file},
+            self_validate_schema => $self->{self_validate_schema},
+            schema_file          => $self->{schema_file}
+        }
+    );
+
+    # Load data in $self
+    $self->{data}              = $data;                  # Dynamically adding attributes (setter)
+    $self->{data_mapping_file} = $data_mapping_file;     # Dynamically adding attributes (setter)
+    $self->{metaData}          = get_metaData($self);    # Dynamically adding attributes (setter)
+    $self->{convertPheno}      = get_info($self);        # Dynamically adding attributes (setter)
+
+    # array_dispatcher will deal with JSON arrays
+    return $self->array_dispatcher;
+}
+
+#############
+#############
+#  CSV2PXF  #
+#############
+#############
+
+sub csv2pxf {
+
+    my $self = shift;
+
+    # First iteration: csv2bff
+    $self->{method} = 'csv2bff';    # setter - we have to change the value of attr {method}
+    my $bff = csv2bff($self);       # array
+
+    # Preparing for second iteration: bff2pxf
+    $self->{method}      = 'bff2pxf';    # setter
+    $self->{data}        = $bff;         # setter
+    $self->{in_textfile} = 0;            # setter
+
+    # Run second iteration
+    return $self->array_dispatcher;
+
 }
 
 #############
@@ -570,8 +692,9 @@ sub pxf2bff {
 
 sub pxf2csv {
 
+    my $self = shift;
     # <array_dispatcher> will deal with JSON arrays
-    return array_dispatcher(shift);
+    return $self->array_dispatcher;
 }
 
 #############
@@ -582,8 +705,22 @@ sub pxf2csv {
 
 sub pxf2jsonf {
 
+    my $self = shift;
     # <array_dispatcher> will deal with JSON arrays
-    return array_dispatcher(shift);
+    return $self->array_dispatcher;
+}
+
+##############
+##############
+# PXF2JSONLD #
+##############
+##############
+
+sub pxf2jsonld {
+
+    my $self = shift;
+    # <array_dispatcher> will deal with JSON arrays
+    return $self->array_dispatcher;
 }
 
 ######################
@@ -598,7 +735,8 @@ sub array_dispatcher {
 
     # Load the input data as Perl data structure
     my $in_data =
-      ( $self->{in_textfile} && $self->{method} !~ m/^redcap2|^omop2|^cdisc2/ )
+      (      $self->{in_textfile}
+          && $self->{method} !~ m/^(redcap2|omop2|cdisc2|csv)/ )
       ? io_yaml_or_json( { filepath => $self->{in_file}, mode => 'read' } )
       : $self->{data};
 
@@ -607,12 +745,16 @@ sub array_dispatcher {
         redcap2bff => \&do_redcap2bff,
         cdisc2bff  => \&do_cdisc2bff,
         omop2bff   => \&do_omop2bff,
+        csv2bff    => \&do_csv2bff,
+        csv2pxf    => \&do_csv2pxf,
         bff2pxf    => \&do_bff2pxf,
         bff2csv    => \&do_bff2csv,
         bff2jsonf  => \&do_bff2csv,      # Not a typo, is the same as above
+        bff2jsonld => \&do_bff2jsonld,
         pxf2bff    => \&do_pxf2bff,
         pxf2csv    => \&do_pxf2csv,
-        pxf2jsonf  => \&do_pxf2csv      # Not a typo, is the same as above
+        pxf2jsonf  => \&do_pxf2csv,      # Not a typo, is the same as above
+        pxf2jsonld => \&do_pxf2jsonld
     );
 
     # Open connection to SQLlite databases ONCE
@@ -625,9 +767,13 @@ sub array_dispatcher {
         say $fh_out "[";
     }
 
-    # Proceed depending if we have an ARRAY or not
-    # NB: Caution with RAM (we store all in memory except for omop2bff)
+    # *** IMPORTANT ***
+    # $out_data = Caution with RAM 
+    # We store all in memory and serialize externally
+    # except for omop2bff (larger) that we print to file here (item-by-item)
     my $out_data;
+
+    # Proceed depending if we have an ARRAY or not
     if ( ref $in_data eq ref [] ) {
 
         # Print if we have ARRAY
@@ -642,15 +788,17 @@ sub array_dispatcher {
         # In $self->{data} we have all participants data, but,
         # WE DELIBERATELY SEPARATE ARRAY ELEMENTS FROM $self->{data}
 
-        for ( @{$in_data} ) {
+        # We void items in $in_data to avoid data duplication in RAM
+        while ( my $item = shift @{ $in_data} ) {         # We want to keep order (!pop)
+
             $count++;
 
-            # Print imfo
+            # Print info
             say "[$count] ARRAY ELEMENT from $elements" if $self->{debug};
 
             # NB: If we get "null" participants the validator will complain
             # about not having "id" or any other required property
-            my $method_result = $func{ $self->{method} }->( $self, $_ );    # Method
+            my $method_result = $func{ $self->{method} }->( $self, $item );    # Method
 
             # Only proceeding if we got value from method
             if ($method_result) {
@@ -661,16 +809,12 @@ sub array_dispatcher {
                 if ( exists $self->{omop_cli} && $self->{omop_cli} ) {
                     my $out = omop_dispatcher( $self, $method_result );
                     print $fh_out $$out;
-                    print $fh_out ",\n"
-                      unless ( $total == $elements
-                        || $total == $self->{max_lines_sql} );
+                    print $fh_out ",\n" unless ( $total == $elements || $total == $self->{max_lines_sql} );
                 }
 
                 # For the other we have array_ref $out_data and serialize at once
                 else {
                     push @{$out_data}, $method_result;
-
-                    #say total_size($out_data);
                 }
             }
         }
@@ -708,13 +852,16 @@ sub omop_dispatcher {
 
     # omop2bff encode directly
     if ( $self->{method_ori} ne 'omop2pxf' ) {
-        $out = JSON::XS->new->utf8->canonical->pretty->encode($method_result);
+
+        # Watch out!! Don't double encode by using utf8 here. $fh is already utf-8!!!
+        #out = JSON::XS->new->utf8->canonical->pretty->encode($method_result);
+        $out = JSON::XS->new->canonical->pretty->encode($method_result);
     }
 
     # omop2pxf convert to PXF
     else {
         my $pxf = do_bff2pxf( $self, $method_result );
-        $out = JSON::XS->new->utf8->canonical->pretty->encode($pxf);
+        $out = JSON::XS->new->canonical->pretty->encode($pxf);
     }
     chomp $out;
     return \$out;
@@ -731,26 +878,16 @@ sub omop_stream_dispatcher {
     # Open a SQLite database connection if required
     open_connections_SQLite($self) if $self->{method} ne 'bff2pxf';
 
-    # Transform Array of Hashes (AoH) to Hash of Hashes (HoH) for faster computation
-    my $person = transform_aoh_to_hoh($self);
-
     # Process files based on the input type (CSV or PostgreSQL dump)
     return @$filepaths
-      ? process_csv_files( $self, $filepaths, $person )
-      : process_sqldump( $self, $filepath, $omop_tables, $person );
+      ? process_csv_files_stream( $self, $filepaths )
+      : process_sqldump_stream( $self, $filepath, $omop_tables, $self );
 }
 
-sub transform_aoh_to_hoh {
+sub process_csv_files_stream {
 
-    my $self   = shift;
-    my $person = { map { $_->{person_id} => $_ } @{ $self->{data}{PERSON} } };
-    delete $self->{data}{PERSON};    # Free up memory
-    return $person;
-}
-
-sub process_csv_files {
-
-    my ( $self, $filepaths, $person ) = @_;
+    my ( $self, $filepaths ) = @_;
+    my $person = $self->{person};
     for my $file (@$filepaths) {
         say "Processing file ... <$file>" if $self->{verbose};
         read_csv_stream(
@@ -765,12 +902,17 @@ sub process_csv_files {
     return 1;
 }
 
-sub process_sqldump {
+sub process_sqldump_stream {
 
-    my ( $self, $filepath, $omop_tables, $person ) = @_;
+    my ( $self, $filepath, $omop_tables ) = @_;
+    my $person = $self->{person};
+
+    # *** IMPORTANT ***
+    # We proceed as we do with CSV, opening the file for every table
+    # With PosgtreSQL.dumps gzipped the overhead means 2x time.
     for my $table (@$omop_tables) {
         next if any { $_ eq $table } @stream_ram_memory_tables;
-        say "Processing table ... <$table>" if $self->{verbose};
+        say "Processing table <$table> line-by-line..." if $self->{verbose};
         $self->{omop_tables} = [$table];
         read_sqldump_stream(
             { in => $filepath, self => $self, person => $person } );
